import datetime
import os
from _decimal import Decimal
from abc import ABC
from enum import Enum
from pathlib import Path
from typing import Any, Dict

from async_lru import alru_cache
from beanie import Document
from pydantic import BaseModel, ConfigDict

from sirius.constants import EnvironmentVariableKey
from sirius.exceptions import ApplicationException, SDKClientException


class Environment(Enum):
    Production = "Production"
    Test = "Test"
    Development = "Development"


class Currency(Enum):
    AED = "AED"
    AUD = "AUD"
    BDT = "BDT"
    BGN = "BGN"
    CAD = "CAD"
    CHF = "CHF"
    CLP = "CLP"
    CNY = "CNY"
    CRC = "CRC"
    CZK = "CZK"
    DKK = "DKK"
    EGP = "EGP"
    EUR = "EUR"
    GBP = "GBP"
    GEL = "GEL"
    HKD = "HKD"
    HUF = "HUF"
    IDR = "IDR"
    ILS = "ILS"
    INR = "INR"
    JPY = "JPY"
    KES = "KES"
    KRW = "KRW"
    LKR = "LKR"
    MAD = "MAD"
    MXN = "MXN"
    MYR = "MYR"
    NGN = "NGN"
    NOK = "NOK"
    NPR = "NPR"
    NZD = "NZD"
    PHP = "PHP"
    PKR = "PKR"
    PLN = "PLN"
    RON = "RON"
    SEK = "SEK"
    SGD = "SGD"
    THB = "THB"
    TRY = "TRY"
    TZS = "TZS"
    UAH = "UAH"
    UGX = "UGX"
    USD = "USD"
    UYU = "UYU"
    VND = "VND"
    XOF = "XOF"
    ZAR = "ZAR"


class DataClass(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ImmutableDataClass(DataClass):
    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)


class PersistedDataClass(Document, ABC):
    id: str  # type:ignore[assignment]


class EnvironmentSecret(PersistedDataClass):
    value: str


@alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
async def get_environmental_secret(key: Enum, default_value: str | None = None) -> str:
    if not isinstance(key, Enum):
        raise SDKClientException("The key has to be an Enum")

    environmental_secret: EnvironmentSecret | None = await EnvironmentSecret.get(key.value)

    if not environmental_secret:
        if default_value:
            environmental_secret = await EnvironmentSecret(id=key.value, value=default_value).save()
        elif not is_test_environment():
            await EnvironmentSecret(id=key.value, value=f"PLACEHOLDER_{key.value}").save()
            raise ApplicationException(f"The environment secret with the following key is not set: {key.value}")
        else:
            return "PLACEHOLDER"

    return environmental_secret.value


def get_environment_variable(key: EnvironmentVariableKey, default_value: str | None = None) -> str:
    value: str | None = os.getenv(key.value)
    if not value and not default_value:
        raise ApplicationException(f"The environment variable with the following key is not set: {key.value}")

    return value if value else default_value


def get_environment() -> Environment:
    try:
        return Environment(get_environment_variable(EnvironmentVariableKey.ENVIRONMENT, "Test"))
    except (ApplicationException, ValueError):  # ValueError is needed because GitHub CI/CD Pipeline has its own Environment Variable
        return Environment.Test


def is_production_environment() -> bool:
    return Environment.Production == get_environment()


def is_test_environment() -> bool:
    return Environment.Test == get_environment()


def is_development_environment() -> bool:
    return Environment.Development == get_environment()


def get_application_name() -> str:
    try:
        return get_environment_variable(EnvironmentVariableKey.APPLICATION_NAME)
    except ApplicationException:
        return Path.cwd().name.title()


def is_dict_include_another_dict(one_dict: Dict[Any, Any], another_dict: Dict[Any, Any]) -> bool:
    if not all(key in one_dict for key in another_dict):
        return False

    for key, value in one_dict.items():
        if another_dict[key] != value:
            return False

    return True


def get_decimal_str(decimal: Decimal) -> str:
    return "{:,.2f}".format(float(decimal))


def get_date_string(date: datetime.date) -> str:
    return date.strftime("%d/%b/%Y")


def get_previous_business_day(datestamp: datetime.date) -> datetime.date:
    previous_business_day: datetime.date = datestamp - datetime.timedelta(days=1)
    return get_previous_business_day_adjusted_date(previous_business_day)


def get_next_business_day_adjusted_date(datestamp: datetime.date) -> datetime.date:
    next_day: datetime.date = datestamp
    while next_day.weekday() >= 5:
        next_day += datetime.timedelta(days=1)

    return next_day


def get_previous_business_day_adjusted_date(datestamp: datetime.date) -> datetime.date:
    prev_day: datetime.date = datestamp
    while prev_day.weekday() >= 5:
        prev_day -= datetime.timedelta(days=1)

    return prev_day



from enum import Enum


class EnvironmentVariableKey(Enum):
    ENVIRONMENT = "ENVIRONMENT"
    APPLICATION_NAME = "APPLICATION_NAME"
    DATABASE_CONNECTION_STRING = "DATABASE_CONNECTION_STRING"
    GITHUB_TOKEN = "GITHUB_TOKEN"
    SUBDOMAIN = "SUBDOMAIN"


class EnvironmentSecretKey:
    pass


class SiriusEnvironmentSecretKey(EnvironmentSecretKey, Enum):
    API_KEY = "API_KEY"
    DISCORD_BOT_TOKEN = "DISCORD_BOT_TOKEN"
    WISE_API_KEY = "WISE_API_KEY"
    IBKR_SERVICE_BASE_URL = "IBKR_SERVICE_BASE_URL"
    ALPHA_VANTAGE_API_KEY = "ALPHA_VANTAGE_API_KEY"
    MASSIVE_API_KEY = "MASSIVE_API_KEY"
    MICROSOFT_CLIENT_ID = "MICROSOFT_CLIENT_ID"
    MICROSOFT_TENANT_ID = "MICROSOFT_TENANT_ID"
    MICROSOFT_CLIENT_SECRET = "MICROSOFT_CLIENT_SECRET"
    MICROSOFT_REDIRECT_URI = "MICROSOFT_REDIRECT_URI"
    LOGIN_REDIRECT_URI = "LOGIN_REDIRECT_URI"
    PRIVATE_KEY = "PRIVATE_KEY"



import json
from abc import abstractmethod
from dataclasses import dataclass
from json import JSONDecodeError
from typing import Dict, Any, List, cast, Callable

import httpx
from httpx import Response, Cookies, Headers, URL, AsyncClient
from httpx._types import RequestFiles
from pydantic import BaseModel

from sirius import common
from sirius.common import DataClass
from sirius.http_requests.exceptions import ClientSideException, ServerSideException


@dataclass
class HTTPResponse:
    response: Response
    response_code: int
    is_successful: bool
    headers: Headers
    data: Dict[Any, Any] | None = None
    response_text: str | None = None
    cookies: Cookies | None = None

    def __init__(self, response: Response, *args: List[Any], **kwargs: Dict[str, Any]) -> None:
        self.response = response
        self.response_code = self.response.status_code
        self.is_successful = 200 <= self.response_code < 300
        self.response_text = self.response.text
        self.headers = response.headers
        self.cookies = response.cookies

        if self.response_text is not None and self.response_text != "":
            try:
                self.data = self.response.json()
            except JSONDecodeError:
                self.data = None

        super().__init__(*args, **kwargs)


class HTTPSession:
    host: str

    @staticmethod
    def raise_http_exception(http_response: HTTPResponse) -> None:
        error_message: str = f"HTTP Exception\n" \
                             f"URL: {str(http_response.response.url)}\n" \
                             f"Headers: {str(http_response.headers)}\n" \
                             f"Method: {http_response.response.request.method.upper()}\n" \
                             f"Response Code: {http_response.response_code}\n" \
                             f"Response Text: {http_response.response_text}"

        if 400 <= http_response.response_code < 500:
            raise ClientSideException(error_message, data={"http_response": http_response})
        else:
            raise ServerSideException(error_message, data={"http_response": http_response})


class AsyncHTTPSession(HTTPSession):
    client: AsyncClient
    _instance_list: List["AsyncHTTPSession"] = []

    def __new__(cls, url_str: str, headers: Dict[str, Any] | None = None, connection_timeout: float = 5.0,
                read_timeout: float = 60.0) -> "AsyncHTTPSession":
        host: str = URL(url_str).host
        instance: AsyncHTTPSession | None = None
        create_client: Callable[[], httpx.AsyncClient] = lambda: httpx.AsyncClient(
            timeout=httpx.Timeout(connect=5.0, read=read_timeout, write=5.0, pool=5.0))

        for i in cls._instance_list:
            if i.host == host and (headers is None or common.is_dict_include_another_dict(cast(Dict[str, Any], headers), dict(i.client.headers))):
                instance = i

        if instance is None:
            instance = super().__new__(cls)
            instance.host = host
            instance.client = create_client()

            if headers is not None:
                instance.client.headers.update(headers)

            cls._instance_list.append(instance)

        if instance.client.is_closed:
            instance.client = create_client()

        return instance

    async def get(self, url: str, query_params: Dict[str, Any] | None = None, headers: Dict[str, Any] | None = None) -> HTTPResponse:
        http_response: HTTPResponse = HTTPResponse(await self.client.get(url, params=query_params, headers=headers))
        if not http_response.is_successful:
            AsyncHTTPSession.raise_http_exception(http_response)

        return http_response

    async def put(self, url: str, data: Dict[str, Any], headers: Dict[str, Any] | None = None, files: RequestFiles | None = None) -> HTTPResponse:
        headers = {} if not headers else headers
        headers.setdefault("content-type", "application/json")
        data_string: str | None = json.dumps(data) if data else None
        http_response: HTTPResponse = HTTPResponse(await self.client.put(url, data=data_string, headers=headers, files=files))  # type: ignore[arg-type]

        if not http_response.is_successful:
            AsyncHTTPSession.raise_http_exception(http_response)

        return http_response

    async def post(self, url: str, data: Dict[str, Any] | str | None = None, headers: Dict[str, Any] | None = None, files: RequestFiles | None = None) -> HTTPResponse:
        headers = {} if not headers else headers

        if files:
            http_response = HTTPResponse(await self.client.post(url, data=data, files=files, headers=headers))  # type: ignore[arg-type]
        else:
            if data and isinstance(data, Dict):
                headers.setdefault("content-type", "application/json")
                data = json.dumps(data)

            http_response = HTTPResponse(await self.client.post(url, data=data, files=files, headers=headers))  # type: ignore[arg-type]

        if not http_response.is_successful:
            self.raise_http_exception(http_response)

        return http_response

    async def delete(self, url: str, headers: Dict[str, Any] | None = None) -> HTTPResponse:
        http_response: HTTPResponse = HTTPResponse(await self.client.delete(url, headers=headers))
        if not http_response.is_successful:
            AsyncHTTPSession.raise_http_exception(http_response)

        return http_response


class HTTPModel(DataClass):

    @abstractmethod
    def __init__(self, **data: Any):
        super().__init__(**data)

    @staticmethod
    async def get_one(cls: type, url: str, query_params: Dict[str, Any] | None = None, http_session: AsyncHTTPSession | None = None) -> DataClass:
        if http_session is None:
            http_session = AsyncHTTPSession(url)

        if not issubclass(cls, BaseModel):
            raise ServerSideException(f"{cls.__name__} is not a Pydantic subclass")

        response: HTTPResponse = await http_session.get(url=url, query_params=query_params)
        return cls(**response.data)  # type: ignore[return-value]

    @staticmethod
    async def get_multiple(cls: type, url: str, query_params: Dict[str, Any] | None = None, headers: Dict[str, Any] | None = None, http_session: AsyncHTTPSession | None = None) -> List[DataClass]:
        if http_session is None:
            http_session = AsyncHTTPSession(url)

        if not issubclass(cls, BaseModel):
            raise ServerSideException(f"{cls.__name__} is not a Pydantic subclass")

        response: HTTPResponse = await http_session.get(url=url, query_params=query_params, headers=headers)
        return [cls(**data) for data in response.data]  # type: ignore[misc]

    @staticmethod
    async def post_return_one(cls: type, url: str, data: Dict[Any, Any] | None = None, headers: Dict[str, Any] | None = None, http_session: AsyncHTTPSession | None = None) -> DataClass:
        if http_session is None:
            http_session = AsyncHTTPSession(url)

        if not issubclass(cls, BaseModel):
            raise ServerSideException(f"{cls.__name__} is not a Pydantic subclass")

        response: HTTPResponse = await http_session.post(url=url, data=data, headers=headers)
        return cls(**response.data)  # type: ignore[return-value]



from sirius.exceptions import ApplicationException


class HTTPException(ApplicationException):
    pass


class ClientSideException(HTTPException):
    pass


class ServerSideException(HTTPException):
    pass



import asyncio
import datetime
import itertools
from decimal import Decimal
from typing import List, Dict, Any, cast, Set, Callable, Optional
from zoneinfo import ZoneInfo

import httpx
from async_lru import alru_cache
from pydantic import ConfigDict

from sirius import common
from sirius.common import Currency
from sirius.constants import SiriusEnvironmentSecretKey
from sirius.exceptions import SiriusException, OperationNotSupportedException
from sirius.http_requests import HTTPResponse, ServerSideException, AsyncHTTPSession
from sirius.market_data import Stock, Option, StockMarketData, Exchange, OptionType

OPTIONS_DATE_FORMAT: str = "%b%y"
_session: AsyncHTTPSession | None = None


async def get_base_url() -> str:
    return await common.get_environmental_secret(SiriusEnvironmentSecretKey.IBKR_SERVICE_BASE_URL, "https://ibkr-service:5000/v1/api/")


async def get_session() -> AsyncHTTPSession:
    global _session
    if not _session:
        base_url: str = await common.get_environmental_secret(SiriusEnvironmentSecretKey.IBKR_SERVICE_BASE_URL, "https://ibkr-service:5000/v1/api/")
        _session = AsyncHTTPSession(base_url)
        _session.client = httpx.AsyncClient(verify=False, timeout=60)

    return _session


class IBKRException(SiriusException):
    pass


class IBKRStock(Stock):
    contract_id: int
    model_config = ConfigDict(frozen=True)

    @staticmethod
    @alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
    async def _find(ticker: str) -> Optional[Stock]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        response: HTTPResponse = await session.get(f"{base_url}iserver/secdef/search?symbol={ticker}&secType=STK")
        filtered_list: List[Dict[str, Any]] = list(filter(lambda d: d["description"] in Exchange, response.data))
        contract_id_list: List[int] = [int(data["conid"]) for data in filtered_list]

        if not contract_id_list:
            return None

        if len(contract_id_list) > 1:
            raise IBKRException(f"More than one stock found for the ticker: {ticker}")

        contract_id: int = contract_id_list[0]
        response = await session.get(f"{base_url}iserver/contract/{contract_id}/info")
        return IBKRStock(
            name=response.data["company_name"],
            currency=Currency(response.data["currency"]),
            ticker=response.data["symbol"],
            contract_id=contract_id
        )


class IBKROption(Option):
    contract_id: int

    @staticmethod
    async def _find(option_id: str) -> Optional["Option"]:
        raise OperationNotSupportedException("Finding IBKR Options using the ID is not yet supported")

    @staticmethod
    async def __get_all_expiry_month_list(stock: IBKRStock, number_of_days_to_expiry: int) -> List[datetime.date]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        response: HTTPResponse = await session.get(f"{base_url}iserver/secdef/search?symbol={stock.ticker}&secType=STK")
        data: Dict[str, Any] = next(filter(lambda c: int(c["conid"]) == stock.contract_id, response.data))
        option_data: Dict[str, Any] = next(filter(lambda o: o["secType"] == "OPT", data["sections"]))
        all_expiry_month_str_list: List[str] = option_data["months"].split(";")
        all_expiry_month_list: List[datetime.date] = [datetime.datetime.strptime(expiry_month, OPTIONS_DATE_FORMAT).date() for expiry_month in all_expiry_month_str_list]

        return [date for date in all_expiry_month_list if (date - datetime.datetime.now().date()).days <= number_of_days_to_expiry]

    @staticmethod
    async def __get_for_strike_and_expiry(stock: IBKRStock, expiry_month: datetime.date, strike_price: Decimal) -> List["IBKROption"]:
        option_list: List[IBKROption] = []
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        expiry_month_str: str = expiry_month.strftime(OPTIONS_DATE_FORMAT).upper()
        response: HTTPResponse = await session.get(
            f"{base_url}iserver/secdef/info",
            query_params={
                "conid": stock.contract_id,
                "sectype": "OPT",
                "month": expiry_month_str,
                "strike": float(strike_price)}
        )

        for data in response.data:
            expiry_date: datetime.date = datetime.datetime.strptime(data["maturityDate"], '%Y%m%d').date()
            option_type: OptionType = OptionType.CALL if data["right"] == "C" else OptionType.PUT
            option_id: str = Option._get_id(stock.ticker, expiry_date, strike_price, option_type)

            option_list.append(IBKROption(
                id=option_id,
                contract_id=data["conid"],
                strike_price=strike_price,
                expiry_date=expiry_date,
                type=option_type,
                underlying_stock=stock,
                name=option_id,
                currency=stock.currency
            ))

        return option_list

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 12 hour cache
    async def _find_all_options(ticker: str, number_of_days_to_expiry: int) -> List[Option]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        abstract_stock: Stock = await Stock.get(ticker)
        stock: IBKRStock = cast(IBKRStock, await IBKRStock._get_local_object(abstract_stock))
        option_contract_list: List[Option] = []
        expiry_month_list: List[datetime.date] = await IBKROption.__get_all_expiry_month_list(stock, number_of_days_to_expiry)
        expiry_month_str_list: List[str] = [expiry_month.strftime(OPTIONS_DATE_FORMAT).upper() for expiry_month in expiry_month_list]

        responses: List[HTTPResponse] = await asyncio.gather(*[
            session.get(f"{base_url}iserver/secdef/strikes", query_params={"conid": stock.contract_id, "sectype": "OPT", "month": expiry_month_str})
            for expiry_month_str in expiry_month_str_list
        ])
        for expiry_month, response in zip(expiry_month_list, responses):
            all_strike_price_set: Set[Decimal] = set([Decimal(str(strike_price)) for strike_price in response.data.get("call", [])])
            all_strike_price_set.update([Decimal(str(strike_price)) for strike_price in response.data.get("put", [])])
            all_option_contract_list: List[IBKROption] = list(itertools.chain.from_iterable(await asyncio.gather(*[
                IBKROption.__get_for_strike_and_expiry(stock, expiry_month, strike_price)
                for strike_price in all_strike_price_set
            ])))

            option_contract_list.extend(
                [option
                 for option in all_option_contract_list
                 if (option.expiry_date - datetime.datetime.now().date()).days == number_of_days_to_expiry]
            )

        return option_contract_list


class IBKRStockMarketData(StockMarketData):

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 12 hour cache
    async def _get_raw_ohlc_data(contract_id: int, from_timestamp: datetime.datetime, to_timestamp: datetime.datetime) -> List[Dict[str, Any]]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        DATE_FORMAT: str = "%Y%m%d-%H:%M:%S"
        try:
            response = await session.get(
                f"{base_url}iserver/marketdata/history",
                query_params={
                    "conid": contract_id,
                    "period": "999d",
                    "bar": "1d",
                    "startTime": (to_timestamp + datetime.timedelta(days=1)).strftime(DATE_FORMAT),  # IBKR sends data 1 day earlier, no idea why.
                    "direction": "-1"
                }
            )
        except ServerSideException as e:
            raise ServerSideException("Did not retrieve any historical market data due to: " + str(e))

        data = response.data.get("data", [])
        if not data:
            return []

        earliest_data_timestamp = min([datetime.datetime.fromtimestamp(d["t"] / 1000) for d in data])

        if earliest_data_timestamp.timestamp() > from_timestamp.timestamp():
            more_data = await IBKRStockMarketData._get_raw_ohlc_data(contract_id, from_timestamp, earliest_data_timestamp)
            return list(itertools.chain(data, more_data))

        return data

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 60 second cache
    async def _get_raw_latest_ohlc_data(contract_id: int) -> Dict[str, Any]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        is_response_valid: Callable = lambda r: "31" in r.data[0].keys() and r.data[0]["31"] != ""
        number_of_tries: int = 1
        response: HTTPResponse = await session.get(f"{base_url}iserver/marketdata/snapshot", query_params={"conids": contract_id, "fields": "7295,70,71,31,87"})
        while number_of_tries < 5 and not is_response_valid(response):
            await asyncio.sleep(0.1)
            response = await session.get(f"{base_url}iserver/marketdata/snapshot", query_params={"conids": contract_id, "fields": "7295,70,71,31,87"})
            number_of_tries = number_of_tries + 1

        if not is_response_valid(response):
            raise ServerSideException("Did not retrieve any market data.")

        return response.data[0]

    @staticmethod
    async def _get(abstract_stock: Stock, from_datestamp: datetime.date, to_datestamp: datetime.date) -> List["StockMarketData"]:
        from_datestamp = common.get_next_business_day_adjusted_date(from_datestamp)
        to_datestamp = common.get_previous_business_day_adjusted_date(to_datestamp)
        from_timestamp: datetime.datetime = datetime.datetime.combine(from_datestamp, datetime.time.min, tzinfo=ZoneInfo("America/New_York"))
        to_timestamp: datetime.datetime = datetime.datetime.combine(to_datestamp, datetime.time.min, tzinfo=ZoneInfo("America/New_York"))
        stock = cast(IBKRStock, await IBKRStock._get_local_object(abstract_stock))
        raw_ohlc_data_list: List[Dict[str, float]] = await IBKRStockMarketData._get_raw_ohlc_data(stock.contract_id, from_timestamp, to_timestamp)

        return [IBKRStockMarketData(
            open=Decimal(str(ohlc_data["o"])),
            high=Decimal(str(ohlc_data["h"])),
            low=Decimal(str(ohlc_data["l"])),
            close=Decimal(str(ohlc_data["c"])),
            timestamp=datetime.datetime.fromtimestamp(ohlc_data["t"] / 1000),
            stock=stock)
            for ohlc_data in raw_ohlc_data_list
            if from_timestamp.timestamp() <= ohlc_data["t"] / 1000 <= to_timestamp.timestamp()]

    @staticmethod
    async def get_latest(abstract_stock: Stock) -> "IBKRStockMarketData":
        ibkr_stock: IBKRStock = cast(IBKRStock, await IBKRStock._get_local_object(abstract_stock))
        raw_data: Dict[str, Any] = await IBKRStockMarketData._get_raw_latest_ohlc_data(ibkr_stock.contract_id)
        open: Decimal = Decimal(raw_data["7295"].replace("H", "").replace("C", ""))
        high: Decimal = Decimal(raw_data["70"].replace("H", "").replace("C", ""))
        low: Decimal = Decimal(raw_data["71"].replace("H", "").replace("C", ""))
        close: Decimal = Decimal(raw_data["31"].replace("H", "").replace("C", ""))
        return IBKRStockMarketData(
            open=open,
            high=high,
            low=low,
            close=close,
            timestamp=datetime.datetime.now(),
            stock=ibkr_stock
        )



from typing import Callable, Dict, Any, Optional

import httpx
import jwt
from async_lru import alru_cache
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from fastapi import Depends, Cookie
from fastapi import Request, Response, status
from fastapi.exceptions import HTTPException
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from jwt import PyJWK
from starlette.responses import JSONResponse

from sirius.common import is_test_environment, get_environmental_secret
from sirius.constants import SiriusEnvironmentSecretKey
from sirius.exceptions import ApplicationException


async def get_private_key() -> str:
    generate_private_key: Callable[[], str] = lambda: rsa.generate_private_key(public_exponent=65537, key_size=2048).private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    ).decode("utf-8")

    if is_test_environment():
        return generate_private_key()

    try:
        return await get_environmental_secret(SiriusEnvironmentSecretKey.PRIVATE_KEY)
    except ApplicationException:
        return await get_environmental_secret(SiriusEnvironmentSecretKey.PRIVATE_KEY, generate_private_key())


@alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
async def get_public_key() -> str:
    private_key_string: str = await get_private_key()
    return (serialization.load_pem_private_key(private_key_string.encode("utf-8"), password=None)
            .public_key()
            .public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo).decode("utf-8"))


@alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
async def get_jwks_client(oidc_config_url: str) -> jwt.PyJWKClient:
    from sirius.http_requests import AsyncHTTPSession
    oidc_config: Dict[str, Any] = (await AsyncHTTPSession(oidc_config_url).get(oidc_config_url)).data
    return jwt.PyJWKClient(oidc_config.get("jwks_uri"))


async def is_token_valid(token: str) -> bool:
    if not token:
        return False

    tenant_id: str = await get_environmental_secret(SiriusEnvironmentSecretKey.MICROSOFT_TENANT_ID)
    audience: str = await get_environmental_secret(SiriusEnvironmentSecretKey.MICROSOFT_CLIENT_ID)
    issuer: str = f"https://sts.windows.net/{tenant_id}/" if is_application_generated_token(token) else f"https://login.microsoftonline.com/{tenant_id}/v2.0"
    oidc_config_url: str = f"https://login.microsoftonline.com/{tenant_id}/v2.0/.well-known/openid-configuration"
    signing_key: PyJWK = (await get_jwks_client(oidc_config_url)).get_signing_key(jwt.get_unverified_header(token)["kid"])

    try:
        jwt.decode(
            token,
            signing_key.key,
            algorithms=["RS256"],
            audience=audience,
            issuer=issuer,
            verify=True
        )
    except Exception:
        return False

    return True


@alru_cache(maxsize=50, ttl=1800)  # 30 minute cache
async def get_access_token() -> str:
    tenant_id: str = await get_environmental_secret(SiriusEnvironmentSecretKey.MICROSOFT_TENANT_ID)
    client_id: str = await get_environmental_secret(SiriusEnvironmentSecretKey.MICROSOFT_CLIENT_ID)
    client_secret: str = await get_environmental_secret(SiriusEnvironmentSecretKey.MICROSOFT_CLIENT_SECRET)

    token_url: str = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"

    payload: Dict[str, str] = {
        "grant_type": "client_credentials",
        "client_id": client_id,
        "client_secret": client_secret,
        "scope": f"{client_id}/.default"
    }

    async with httpx.AsyncClient() as client:
        response: httpx._models.Response = await client.post(token_url, data=payload)
        response.raise_for_status()
        access_token: str = response.json()["access_token"]

    return access_token


async def get_central_finite_curve_authentication_headers() -> Dict[str, str]:
    return {"Authorization": f"Bearer {await get_access_token()}"}


def is_application_generated_token(token: str) -> bool:
    return "sts.windows.net" in jwt.decode(token, options={"verify_signature": False}).get("iss")


async def verify_token(header_token: HTTPAuthorizationCredentials = Depends(HTTPBearer(auto_error=False)), cookie_token: Optional[str] = Cookie(None, alias="Authorization")) -> None:
    token: str | None = cookie_token.replace("Bearer ", "") if cookie_token else header_token.credentials if header_token else None
    if not await is_token_valid(token):
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, headers={"WWW-Authenticate": "Bearer"}, )


async def redirect_unauthorized_handler(request: Request, exc: HTTPException) -> Response:
    return JSONResponse(
        status_code=exc.status_code,
        content={"message": exc.detail},
        headers={"WWW-Authenticate": "Bearer"} if exc.status_code == status.HTTP_401_UNAUTHORIZED else None,
    )



import asyncio
import datetime
from abc import ABC, abstractmethod
from typing import Optional, List, cast, Dict, Set, Any

from async_lru import alru_cache
from beanie import DecimalAnnotation, Link

from sirius import common
from sirius.common import PersistedDataClass
from sirius.exceptions import DataIntegrityException
from sirius.market_data import Stock, StockMarketData, Option, MarketDataException, OptionMarketData


class Cache(ABC):

    @staticmethod
    @abstractmethod
    async def is_update_cache(*args: List[Any]) -> bool:
        ...

    @staticmethod
    @abstractmethod
    async def update_cache(*args: List[Any]) -> Any:
        ...

    @staticmethod
    @abstractmethod
    async def get_from_data_provider(*args: List[Any]) -> Any:
        ...


class CachedStock(PersistedDataClass, Stock, Cache):  # type:ignore[misc]

    @staticmethod
    @alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
    async def get_from_data_provider(ticker: str) -> Optional["CachedStock"]:
        from sirius.market_data.alpha_vantage import AlphaVantageStock
        alpha_vantage_stock: AlphaVantageStock = cast(AlphaVantageStock, await AlphaVantageStock._find(ticker))
        return CachedStock(
            id=alpha_vantage_stock.ticker,
            name=alpha_vantage_stock.name,
            ticker=alpha_vantage_stock.ticker,
            currency=alpha_vantage_stock.currency
        )

    @staticmethod
    @alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
    async def _find(ticker: str) -> Optional["Stock"]:
        cached_stock: CachedStock | None = await CachedStock.get(ticker)
        if cached_stock:
            return cached_stock

        if await CachedStock.is_update_cache(ticker):
            return await CachedStock.update_cache(ticker)

        raise DataIntegrityException(f"Cached stock could not be found in cache nor is supposed to be retrieved from the cached. Ticker: {ticker}")  # This is just kept for best practice.

    @staticmethod
    async def is_update_cache(ticker: str) -> bool:  # type: ignore[override]
        return await CachedStock.get(ticker) is None

    @staticmethod
    async def update_cache(ticker: str) -> Stock:  # type: ignore[override]
        stock = await CachedStock.get_from_data_provider(ticker)
        if not stock:
            raise MarketDataException(f"Could not find stock from data provider with ticker: {ticker}")

        return await stock.save()


class CachedStockMarketData(PersistedDataClass, StockMarketData, Cache):  # type:ignore[misc]
    open: DecimalAnnotation
    high: DecimalAnnotation
    low: DecimalAnnotation
    close: DecimalAnnotation
    stock: Link[CachedStock]  # type: ignore[assignment]

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 12 hour cache
    async def get_from_data_provider(ticker: str, from_timestamp: datetime.datetime) -> List["CachedStockMarketData"]:
        from sirius.market_data.massive import MassiveStockMarketData
        abstract_stock: Stock = await Stock.get(ticker)
        stock: CachedStock = cast(CachedStock, await CachedStock._get_local_object(abstract_stock))
        latest_market_data_list: List[StockMarketData] = await MassiveStockMarketData._get(abstract_stock, from_timestamp, datetime.datetime.now())

        return [CachedStockMarketData(
            id=f"{market_data.stock.ticker} | {int(market_data.timestamp.timestamp())}",
            open=market_data.open,
            high=market_data.high,
            low=market_data.low,
            close=market_data.close,
            timestamp=market_data.timestamp,
            stock=stock)
            for market_data in latest_market_data_list]

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 12 hour cache
    async def _get_all(ticker: str, is_update_cache: bool = True) -> List["CachedStockMarketData"]:
        if not common.is_test_environment():
            all_data_list: List[CachedStockMarketData] = await CachedStockMarketData.find(CachedStockMarketData.stock.id == ticker, fetch_links=False).to_list()  # type: ignore[attr-defined]
        else:
            all_data_list = [c for c in await CachedStockMarketData.find_all().to_list() if c.stock.ref.id == ticker]  # TODO: Add get_underlying_stock()

        if is_update_cache and await CachedStockMarketData.is_update_cache(ticker, all_data_list):
            latest_data: CachedStockMarketData | None = max(all_data_list, key=lambda d: d.timestamp) if all_data_list else None
            all_data_list = all_data_list + await CachedStockMarketData.update_cache(ticker, latest_data.timestamp if latest_data else None)

        return all_data_list

    @staticmethod
    async def _get(abstract_stock: Stock, from_datestamp: datetime.date, to_datestamp: datetime.date) -> List["StockMarketData"]:
        from_datestamp = common.get_next_business_day_adjusted_date(from_datestamp)
        to_datestamp = common.get_previous_business_day_adjusted_date(to_datestamp)
        cached_data_list: List[CachedStockMarketData] = await CachedStockMarketData._get_all(abstract_stock.ticker)

        return [obj for obj in cached_data_list if from_datestamp <= obj.timestamp.date() <= to_datestamp]

    @staticmethod
    async def is_update_cache(ticker: str, all_data_list: List["CachedStockMarketData"] | None = None) -> bool:  # type: ignore[override]
        all_data_list = await CachedStockMarketData.find(CachedStockMarketData.stock.id == ticker, fetch_links=False).to_list() if not all_data_list else all_data_list  # type: ignore[attr-defined]
        latest_data: CachedStockMarketData | None = max(all_data_list, key=lambda d: d.timestamp) if all_data_list else None
        expected_latest_data_date: datetime.date = common.get_previous_business_day(datetime.date.today())

        return len(all_data_list) == 0 or (expected_latest_data_date - latest_data.timestamp.date()).days > 1

    @staticmethod
    async def update_cache(ticker: str, from_timestamp: datetime.datetime | None = None) -> List["CachedStockMarketData"]:  # type: ignore[override]
        from_timestamp = (datetime.datetime.now().replace(year=datetime.datetime.now().year - 10)) if not from_timestamp else from_timestamp  # 10 years
        current_data: Dict[str, CachedStockMarketData] = {data.id: data for data in await CachedStockMarketData._get_all(ticker, is_update_cache=False)}
        latest_data: Dict[str, CachedStockMarketData] = {data.id: data for data in await CachedStockMarketData.get_from_data_provider(ticker, from_timestamp)}
        new_data_ids: Set[str] = latest_data.keys() - current_data.keys()
        unique_data_to_update_list: List[CachedStockMarketData] = [latest_data[k] for k in new_data_ids]
        await CachedStockMarketData.insert_many(unique_data_to_update_list)

        return unique_data_to_update_list


class CachedOption(PersistedDataClass, Option, Cache):
    strike_price: DecimalAnnotation
    underlying_stock: Link[CachedStock]  # type: ignore[assignment]

    async def get_underlying_stock(self) -> Stock:
        if isinstance(self.underlying_stock, Link):
            self.underlying_stock = await self.underlying_stock.fetch()

        return self.underlying_stock  # type: ignore[return-value]

    @staticmethod
    async def _find(option_id: str) -> Optional["Option"]:
        return await CachedOption.get(option_id)

    @staticmethod
    @alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
    async def get_from_data_provider(ticker: str) -> List["CachedOption"]:
        from sirius.market_data.alpha_vantage import AlphaVantageOption
        stock: Stock | None = await Stock.find(ticker)
        if not stock:
            return []

        cached_stock: CachedStock | None = await CachedStock.get(ticker)
        all_option_list: List[AlphaVantageOption] = await AlphaVantageOption._find_all_from_ticker(ticker)
        return [CachedOption(
            id=Option._get_id(ticker, option.expiry_date, option.strike_price, option.type),
            name=Option._get_id(ticker, option.expiry_date, option.strike_price, option.type),
            strike_price=option.strike_price,
            expiry_date=option.expiry_date,
            type=option.type,
            currency=option.currency,
            underlying_stock=cached_stock
        ) for option in all_option_list]

    @staticmethod
    @alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
    async def _find_all_options(ticker: str, number_of_days_to_expiry: int, is_update_cache: bool = True) -> List["Option"]:
        expiry_date: datetime.date = datetime.datetime.now().date() + datetime.timedelta(days=number_of_days_to_expiry)
        if not common.is_test_environment():
            all_options_list: List[CachedOption] = await CachedOption.find(CachedOption.underlying_stock.id == ticker).to_list()  # type: ignore[attr-defined]
        else:
            all_options_list = [c for c in await CachedOption.find_all().to_list() if c.underlying_stock.ref.id == ticker]

        if is_update_cache and await CachedOption.is_update_cache(ticker, all_options_list):
            all_options_list = await CachedOption.update_cache(ticker)

        return [option for option in all_options_list if option.expiry_date == expiry_date]

    @staticmethod
    async def is_update_cache(ticker: str, all_options_list: List["CachedOption"] | None = None) -> bool:  # type: ignore[override]
        all_options_list = await CachedOption.find(CachedOption.underlying_stock.id == ticker).to_list() if not all_options_list else all_options_list  # type: ignore[attr-defined]
        earliest_option: CachedOption = min(all_options_list, key=lambda o: o.expiry_date) if len(all_options_list) > 0 else None

        return len(all_options_list) == 0 or earliest_option.expiry_date < datetime.datetime.now().date()

    @staticmethod
    async def update_cache(ticker: str) -> List["CachedOption"]:  # type: ignore[override]
        new_data_list: List[CachedOption] = await CachedOption.get_from_data_provider(ticker)
        await CachedOption.find(CachedOption.underlying_stock.id == ticker).delete()  # type: ignore[attr-defined]
        await CachedOption.insert_many(new_data_list)
        return new_data_list


class CachedOptionMarketData(PersistedDataClass, OptionMarketData, Cache):
    id: str  # type: ignore[assignment]
    last: DecimalAnnotation
    last_updated: datetime.datetime
    option: Link[CachedOption]  # type: ignore[assignment]

    async def get_option(self) -> Option:
        if isinstance(self.option, Link):
            self.option = await self.option.fetch()

        return self.option  # type: ignore[return-value]

    @staticmethod
    async def _get_latest(abstract_option: Option, is_update_cache: bool = True) -> "CachedOptionMarketData":
        cached_option_market_data: CachedOptionMarketData | None = await CachedOptionMarketData.get(abstract_option.id)

        if is_update_cache and await CachedOptionMarketData.is_update_cache(abstract_option.id, cached_option_market_data):
            cached_option_market_data = await CachedOptionMarketData.update_cache(abstract_option)

        return cached_option_market_data

    @staticmethod
    async def is_update_cache(option_id: str, cached_option_market_data: Optional["CachedOptionMarketData"] = None) -> bool:  # type: ignore[override]
        cached_option_market_data = await CachedOptionMarketData.get(option_id) if not cached_option_market_data else cached_option_market_data
        allowed_oldest_datestamp: datetime.date= common.get_previous_business_day(datetime.date.today())
        return not cached_option_market_data or cached_option_market_data.last_updated.date() < allowed_oldest_datestamp

    @staticmethod
    async def get_from_data_provider(abstract_option: Option) -> "CachedOptionMarketData":  # type: ignore[override]
        from sirius.market_data.alpha_vantage import AlphaVantageOptionMarketData
        alpha_vantage_option: AlphaVantageOptionMarketData = await AlphaVantageOptionMarketData._get_latest(abstract_option)
        return CachedOptionMarketData(
            id=alpha_vantage_option.option.id,
            last=alpha_vantage_option.last,
            last_updated=datetime.datetime.now(),
            option=abstract_option  # type: ignore[arg-type]
        )

    @staticmethod
    async def update_cache(abstract_option: Option) -> "CachedOptionMarketData":  # type: ignore[override]
        if not common.is_test_environment():
            all_cached_option_list: List[CachedOption] = await CachedOption.find(CachedOption.underlying_stock.id == abstract_option.underlying_stock.ticker).to_list()  # type: ignore[attr-defined]
        else:
            all_cached_option_list = [c for c in await CachedOption.find_all().to_list() if c.underlying_stock.ref.id == abstract_option.underlying_stock.ticker]

        cached_option_market_data: List[CachedOptionMarketData] = await asyncio.gather(*[CachedOptionMarketData.get_from_data_provider(option) for option in all_cached_option_list])
        cached_option_market_data = await asyncio.gather(*[cached_market_data.save() for cached_market_data in cached_option_market_data])

        return next(o for o in cached_option_market_data if o.id == abstract_option.id)
